<?php
/**
 * $Id: carbon_calculate.inc,v 1.26 2009/01/06 22:19:26 johnackers Exp $
 *
 * Drupal carbon module (calculations)
 *
 * Created on 18-Oct-2006
 *
 * @file
 * The calculations used to build carbon footprint.
 *
 * Different carbon_stamps have a different scope, some declare
 * emissions on one date, others declare emissions over a period
 * of time.
 *
 * Some sources of emissions e.g. from air travel need special
 * formulas for calculating resulting CO2 because fuel used in
 * aircraft takeoff has to be taken into consideration.
 */


/*
 * Describes the footprint results for a single account.
 *
 * used to pass results information between main functions
 * including the reporting functions and the CRAG settlement functions.
 */

class GroupResult
{
	var $account ;  // account id
	var $group ;    // group name
	var $start ;
	var $end ;

	var $sector_kwh0 = array();
	var $sector_kwh1 = array() ;
	var $sector_co2 = array() ;
	var $grand_kwh0 = 0 ;
	var $grand_kwh1 = 0 ;
	var $grand_co2 = 0 ;
	var $C02_above_target = 0 ;
	var $emissions = array();
}

/**
 * Describes a single emission source e.g. a plane trip
 *
 */

class Emission
{
	var $start ; 		// stamp use for starting point
	var $end ; 			// stamp used for end point
	var $error ;
	var $notes ; 		// additional information to present to user

	var $select ;		  // array of stamps
	var $co2   ;		  // co2 emission
	var $kwh0  ; 	    // energy 0
	var $kwh1  ;	    // energy 1

	function Emission($select)
	{
		$this->select = $select;
	}
}


class Stamp
{

}

/**
 * Create a set of results for a given period (usually a year)
 * based on on all the available carbon_stamps. Maintain
 * sector (e.g. home, tranport) and grand totals.
 *
 * @param $account passed to $results but no attributes used
 * @param $stamps array of all carbon_stamps (with associated carbon_source fields)
 * @param $start first date the generated carbon footprint
 * @param $end last date of the generated carbon footprint
 *
 * @return a complex structure that has the carbon footprint for this period
 */


function carbon_calculate_total(&$account, &$stamps, $start, $end)
{
	$results = new GroupResult();
	$results->account = $account ;
	$results->start = $start ;
	$results->end = $end ;
	$results->emissions = _calculate($stamps, $start, $end);
	 
	foreach ($results->emissions as $emission)
	{
		$sector = $emission->select[0]->sector ;
		if (isset($emission->kwh0))
		{
			//aacc
			if (!isset($results->sector_kwh0[$sector]))
			$results->sector_kwh0[$sector] = 0 ;

			$results->sector_kwh0[$sector] += $emission->kwh0 ;
			$results->grand_kwh0 += $emission->kwh0 ;
		}
		if (isset($emission->kwh1))
		{
			//aacc
			if (!isset($results->sector_kwh1[$sector]))
			$results->sector_kwh1[$sector] = 0 ;

			$results->sector_kwh1[$sector] += $emission->kwh1 ;
			$results->grand_kwh1 += $emission->kwh1 ;
		}
		if (isset($emission->co2))
		{
			if (!isset($results->sector_co2[$sector]))
			$results->sector_co2[$sector] = 0 ;

			$results->sector_co2[$sector] += $emission->co2 ;
			$results->grand_co2 += $emission->co2 ;
		}
		// need to execute this once.
		$results->enablekwh0 = $emission->select[0]->enablekwh0 ;
		$results->enablekwh1 = $emission->select[0]->enablekwh1 ;
		$results->enableco2 =  $emission->select[0]->enableco2 ;
	}
	return $results ;
}



/**
 * Aggregate all similar carbon stamps and build an array
 * of results of CO2 calculations.
 *
 * @param $ar array of all carbon_stamps (with associated carbon_source fields)
 * @param $start first date the generated carbon footprint
 * @param $end last date of the generated carbon footprint
 *
 * @return an array of emissions. Each result contains text and/or value.
 */


function _calculate(&$ar, $start, $stop)
{
	$dgroup = array();
	foreach ($ar as $node)
	{
		$dgroup[] = _concat_same_value($node);
	}

	// each key represents a group of readings
	// that can be considered identical.

	$keys = array_unique($dgroup);
	$emissions = array();
	foreach ($keys as $subgroup)
	{
		// pull out the first group and
		// select all the nodes that match that group

		$select = array();
		$column_notes = "" ;
		foreach ($ar as $node2)
		{
			if (_concat_same_value($node2) != $subgroup)
			continue ;
			$select[] = $node2;
			$node2->total = null ;
		}
		$fname = '_scope_'.$select[0]->scope;

		if (function_exists($fname))
		{
			$fname($emissions, $select, $start, $stop);
		}
		else
		{
			drupal_set_message("Invalid or missing function ". $fname, "error");
		}
	}
	return $emissions ;
}

/**
 * identify stamps that are identical so that they can be
 * group together.
 *
 * meter readings (scope == 1) are handled as a group, so stamps
 * with a different value should be group together.
 *
 * ja 2-jan-2009 use _concat() function
 */


function _concat_same_value($node)
{
	$hashcode = _concat($node);
	switch ($node->scope)
	{
		case 0 :
		case 2 :
			return $hashcode.'.'.$node->reading.'.'.$node->enddate.'.'.$node->startdate;
		case 1 : return $hashcode ;
	}
}

/**
 * identify stamps that are identical that can presented in a group on a form
 * and the only the field that needs to be changed is enddate and reading.
 *
 * ja 2-jan-2009 ignore adjustment and code fields. The side effect is that
 * the title of each meter reading determines which sequence of readings it
 * belongs to.
 */

function _concat($node){
	return $node->title.'.'.$node->scope.'.'.$node->sourceclass;
}


/**
 * todo check that no ; or : present
 */

function _evaluate($expression)
{
	$expression = trim($expression);
	$exp1 = substr($expression,0,1);
	if (strcmp($exp1,"=")==0)
	$expression = substr($expression,1);

	return eval('return '.$expression.';');
}



function _stamp_map_sorted_by_enddate($ar)
{
	$ar1 = array();
	foreach ($ar as $node)
	{
		$k = $node->enddate;
		while (isset($ar1[$k]))
		$k++ ;
		$ar1[$node->enddate] = $node ;
	}
	ksort($ar1);
	return $ar1;
}


function _sort_by_valuedate($ar)
{
	$ar1 = array();
	foreach ($ar as $node)
	{
		$k = $node->valuedate;
		while (isset($ar1[$k]))
		$k++ ;
		$ar1[$node->valuedate] = $node ;
	}
	ksort($ar1);
	return array_values($ar1);
}




/**
 * There are two scenarios.
 * 1. the emissions occur on a single date (e.g. air travel)
 * 2. the emissions are spread over a month , year or more.
 *
 * @return false if no stamp falls outside date range
 */

function _scope_0(&$emissions, $select, $start, $stop)
{
	$emission = new Emission($select);
	_duplicate_check($emission->select);

	$stamp = $emission->select[0] ;
	if (empty($stamp->startdate) || $stamp->startdate == $stamp->enddate)
	{
		// this is the simple case where all the emissions
		// were released  on the same date
		// (take care with < and <= etc to include all stamps in
		// at least one accounting period (a year)
		//
    if ($stamp->enddate < $start) return false ;
    if ($stamp->enddate >= $stop) return false ;		 
		_invoke_calculate($emission, $stamp, 1);
	}
	else
	{
		// this is the case where the emissions spread across a long
		// period of time e.g. the carbon stamp might refer to emissions
		// from june 05 to june 06 and so would appear in two separate years.
		 
		// explanation of below:
		// Let's say user inputs carbon stamp of 1000 Kg of CO2 over 10 years.
		// We want to allocate the 1000Kg over the ten years as evenly as possible.
		// So we calculate an adjustment which equals
		//    number of emitting days inside our period of interest (i.e between $start and $end) /
		//    total number of emitting days.
		
    if ($stamp->startdate <= $end) return false ;
		
		$begin = max($stamp->startdate, $start);
		$end = min($stamp->enddate, $stop);
		if ($end <= $begin) return false ;
		 
		$adjust_for_period = ($end - $begin) / ($stamp->enddate - $stamp->startdate);
		_invoke_calculate($emission, $stamp, $adjust_for_period);

	}
	$emission->valuedate = $stamp->enddate ;

	$emissions[] = $emission ;
}

/**
 * meter readings
 *
 * @return false if no stamp falls outside date range
 *
 */


function _scope_1(&$emissions, &$select, $start, $end)
{
	$emission = new Emission($select);

	$num = sizeof($emission->select);

	if ($num == 0)
	{
		$emission->error = "No readings" ;
		return false ;
	}

	$meter_readings = _stamp_map_sorted_by_enddate($emission->select);
	$meter_readings_vals = array_values($meter_readings);

	// do not attempt to make sense of requests for
	// consumption dates that fall entirely below or entirely
	// above the range of available meter readings
	 
	if ($end <= $meter_readings_vals[0]->enddate)
	return ;

	if ($start > $meter_readings_vals[count($meter_readings_vals)-1]->enddate)
	return ;

	$i = _index_of_closest($meter_readings, $start);
	$j = _index_of_closest($meter_readings, $end);

	// extract the meter readings that fall roughly between
	// the $start and and $end times that we are interested in.

	$meter_readings_sliced = array_slice($meter_readings, $i, $j-$i+1);

	// chack that we have enough readings

	if (count($meter_readings_sliced) == 1)
	{
		$emission->error = "Can only find one suitable meter reading" ;
		$emissions[] = $emission ;   return ;
	}

	if (count($meter_readings_sliced) == 0)
	{
		$emission->error = "No suitable meter readings" ;
		$emissions[] = $emission ;   return ;
	}

	// People move in and out of house shares all the time
	// so the 'adjustment' may change several times within
	// a six month or twelve month period.

	//  $meter_last = end($meter_readings_sliced);
	//  $meter_first = reset($meter_readings_sliced);

	$s0 = null ;  $k = 0 ;
	foreach ($meter_readings_sliced as $current)
	{
    $e0 = php4_clone ($emission);
     
    // skip the first iteration
    $e0->start = $s0 ;
    $e0->end = $s0 = $current ;
    if ($e0->start == null)
    continue ;
     
    // boundary adjustments
     
    $adjust_for_start_period = $adjust_for_end_period = 1.0 ;
     
    if ($k++ == 0)  // first meter reading
    {
      $adjust_for_start_period *=  ($e0->end->enddate - $start) / ($e0->end->enddate - $e0->start->enddate);
    }
    if ($k == $j - $i) // last meter reading, must be better way of determining last in array
    {
      $adjust_for_end_period *=  ($end - $e0->start->enddate) / ($e0->end->enddate - $e0->start->enddate);
    }

    $consumption = $e0->end->reading - $e0->start->reading ;
    if ($consumption < 0)
    {
      $e0->error = "Meter appears to have gone backwards ".$consumption ;
      $emissions[] = $e0 ;   return ;
    }


    // but we now need to
    // calculate an adjustment which will scale the value read
    // between meter readings into the period for which we are
    // about to report.
     
    // create an artificial 'one off', scope 0 carbon stamp
     
    $stamp = new Stamp();
    $stamp->sourceclass = $e0->end->sourceclass ;
    $stamp->sourcecode = $e0->end->sourcecode ;
    $stamp->sourcestatus = $e0->end->sourcestatus ;
    $stamp->adjustment = $e0->end->adjustment ;
    $stamp->tokwh0 = $e0->end->tokwh0 ;
    $stamp->tokwh1 = $e0->end->tokwh1 ;
    $stamp->toco2 = $e0->end->toco2 ;
    $stamp->reading = $consumption;
     
    _invoke_calculate($e0, $stamp, $adjust_for_start_period * $adjust_for_end_period) ;

    $e0->notes = "Using readings ".
    $e0->start->reading. " on " . substr(_format_date($e0->start->enddate, "small"),0,10).
       " and ". 
    $e0->end->reading. " on " . substr(_format_date($e0->end->enddate, "small"),0,10) ;
     
    if ($e0->end->adjustment != "1")
    $e0->notes .= " adjusted by ". $e0->end->adjustment  ;
     
    $e0->notes .= ". "  ;
     
    if ($adjust_for_start_period < 0.95 || $adjust_for_start_period > 1.05)
    $e0->notes .= " Multiplied by ". sprintf("%.1f", $adjust_for_start_period). " because ".
        "nearest meter reading not close to start date of accounting period";

    if ($adjust_for_end_period < 0.95 || $adjust_for_end_period > 1.05)
    $e0->notes .= " Muliplied by ". sprintf("%.1f", $adjust_for_end_period). " because ".
        "nearest meter reading not close to end date of accounting period";        

    $e0->valuedate = $e0->end->enddate;

    $emissions[] = $e0;
  }
}

/**
 * daily event
 * The difference betweeen a scope_0 and scope_2 event is that
 * the scope_2 event repeats the same level of emissions every
 * day and the emissions are not dispersed between start and
 * date.
 */


function _scope_2(&$emissions, &$select, $start, $stop)
{
  $emission = new Emission($select);

  _duplicate_check($emission->select);

  $stamp = $emission->select[0] ;
   
  if (empty($stamp->startdate))
  {
    if ($stamp->enddate > $start && $stamp->enddate <= $stop)
      $begin = $start ;
    else
      return false;
  }
  else
  {
    // find out the range of dates when this event occurred
    // and that fall inside the reporting period that we are
    // interested in.

    $begin = max($stamp->startdate, $start);
  }
  $end = min($stamp->enddate, $stop);

  if ($end <= $begin) 
    return false ;
   
  $adjust_for_repetition = ($end - $begin) / (60*60*24);
  _invoke_calculate($emission, $stamp, $adjust_for_repetition);
  $emission->valuedate = $end;

  $emissions[] = $emission;   // process only 1 stamp, ignore duplicates
}

/**
 * We had expected to be passed an array of just 1 carbon_stamp.
 * Log an error if there is more than one.
 * @param $ar an array of carbon stamps
 */


function _duplicate_check($ar)
{
  if (count($ar) > 1)
  {
    $text = "These identical carbon stamps treated as one:";
    foreach ($ar as $stamp)
    {
      $text .= " ".l("node ".$stamp->nid,"/node/".$stamp->nid);
    }
    drupal_set_message($text, "warning");
  }
}


/**
 * walk through the array and look for carbon_stamps that are
 * closest to the required date - used to find meter readings
 * near the start and end dates of a carbon footprint.
 *
 * @param $ar array of all carbon_stamps
 */

function _index_of_closest($map, $date)
{
  $node_found = null ;
  $i = 0 ;
  foreach ($map as $date1 => $node)
  {
    $d = $node->enddate - $date ;
    if ($d < 0) $d = -$d ;
    if (!isset($diff_found) || $d < $diff_found)
    {
      $diff_found = $d ; $node_found = $node ; $i_found = $i ;
    }
    $i++ ;
  }
  return $i_found ;
}


/**
 * Each carbon source can have a different classname.
 * Invoked from carbon_source.inc
 *
 * @return a map of class names (keys) and descriptions (values). Descriptions (values) not used yet.
 */

function carbon_calculate_class_map()
{
  return array(
    "simple"        =>  t("simple multiplier"),
    "climatecare" =>    t("model used by commercial offsetting company"),
    "chooseclimate" =>  t("model created by climate mathematician")
  );
}


/**
 *
 * At the moment the classname is simply mapped to a function name.
 *
 * @param $emission structure into which results are place
 * @param $node a carbon stamp
 */


function _invoke_calculate(&$emission, $node, $adjust)
{
  if (!isset($node->sourcecode))
  {
    $emission->sourceError = "Unknown source code: ".$node->code ;
    return ;
  }
  if ($node->sourcestatus == 0)
  {
    $emission->sourceError = "Source not or no longer published: ".$node->code ;
    return ;
  }
  $fn = '_'.$node->sourceclass ;
  $fn($emission, $node, $adjust);
}

/**
 * The simple class calculates emissions with the formula
 *
 * emissions = r*m*a
 *
 * where r is the value spcified by the user
 *       m is the multiplier used to convert units of r
 *       a is the adjustment specified by the user
 *       which might be 48/52 to allow for holidays
 *       from a daily commute.
 *       (this can be a string and must be evaluated)
 *
 * to the corresponding CO2 value.
 *
 * @param $node a carbon stamp
 */

function _simple(&$emission, $node, $adjust)
{
  $emission->units = $node->reading * _evaluate($node->adjustment) * $adjust;
  $emission->co2 =  $emission->units * $node->toco2 ;
  $emission->kwh0 = $emission->units * $node->tokwh0 ;
  $emission->kwh1 = $emission->units * $node->tokwh1 ;
}

/**
 *  This formula comes from http://www.chooseclimate.org
 *

 Fuel per passenger (Data for B-747).

 Fuel = [7840 + 10.1 * (distance-250)] (*2 if return)
 (7840 kg take off-climb-descent, 10.1 kg/km cruising.
 Passengers, = 370 * [occupancy] (/1.5 if business)
 Greenhouse warming:
 CO2 = fuel * (44/12 * 156/184) (molecular masses)

 Total warming effect of CO2,Ozone (made by NOx), water vapour and
 contrails is about three times greater than effect of CO2 alone
 */


/**
 * $multiplier should only be used to convert distance from some
 * arbitrary unit to Km.
 **/

function _chooseclimate(&$emission, $node, $adjust)
{
  $model = array(
    "aircraft" => "B747-400",
    "fuel_for_takeoff" => 7840,
    "fuel_cruising_per_km" => 10.1,    
    "climb_distance" => 250,  
    "capacity" => 370,  
    "occupancy" => .80,
    "fuel_conversion" => (44/12 * 156/184),
    "radiative_forcing_index" => 2.7
  );
  return _flight_calculate($emission, $node, $adjust, $model);
}

/**
 Climate care model

 This model is based on the information in a report written
 by the Environmnetal Change Institute for climate care
 and is provided on the climate care website.
 http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf

 the raw data refrred to on aircraft fuel consumption
 http://reports.eea.europa.eu/EMEPCORINAIR3/en/B851vs2.4.pdf

 my version of this with average fuel consumption added:
 http://cvs.drupal.org/viewcvs/ checkout /drupal/contributions/modules/carbon/air_fuel_use.html

 note the climate care assume 100% occupancy and that on long
 haul flights the 10% of fuel can be be attributed to freight.
 */

function _climatecare(&$emission, $node, $adjust)
{
  $distance_km = $node->reading * $node->toco2 ;

  // switch planes depending on distance

  $model =  $distance_km < 3500 ? _climatecareB737() : _climatecareB747();

  return _flight_calculate($emission, $node, $adjust, $model);
}

function _climatecareB747() {
  return array(
    "aircraft" => "B747-400",
    "fuel_for_takeoff" => 3402,
    "fuel_cruising_per_km" => 10.1,    // from average kg/mile in EU table
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 400,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // ditto & http://iacweb.ethz.ch/tradeoff/
  );
}

function _climatecareB737() {
  return array(
    "aircraft" => "B737-400",
    "fuel_for_takeoff" => 825.4,
    "fuel_cruising_per_km" => 3.01,    // from average kg/mile above
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 180,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // diito & http://iacweb.ethz.ch/tradeoff/
  );
}

function _climatecareA340() {
  return array(
    "aircraft" => "Airbus-A340",
    "fuel_for_takeoff" => 2019,
    "fuel_cruising_per_km" => 6.6,    // from average kg/mile in EU table
    "climb_distance" => 0,  
    "fuel_conversion" => 3.15,         // http://www.climatecare.org/_media/documents/pdf/Aviation_Emissions_&_Offsets.pdf
    "capacity" => 295,                 // ditto
    "occupancy" => 1.00,               // ditto
    "radiative_forcing_index" => 2.0   // diito & http://iacweb.ethz.ch/tradeoff/
  );
}


/**
 * the choose climate model is basically a simple
 * first order polynomial and I have used it for
 * the climate care model as well.
 */


function _flight_calculate(&$emission, $node, $adjust, $model)
{
  $distance = $node->reading * $node->toco2 ;

  $flight_takeoff = $model["fuel_for_takeoff"] ;

  if ($distance > $model["climb_distance"])
  $flight_cruising += ($model["fuel_cruising_per_km"] * ($distance-$model["climb_distance"]));  // Kg

  $fuel = ($flight_takeoff + $flight_cruising)/($model["capacity"] * $model["occupancy"]) ;   //.Kg

  $co2 = $fuel * $model["fuel_conversion"];  // convert fuel to CO2

  // the radiative_forcing_index below (expected to be between
  // 1.8 and around 2.7) includes effect of the non CO2 gasses

  $co2_and_other_gasses = $co2 * $model["radiative_forcing_index"] ;

  $emission->notes =  sprintf("Assumptions: aircraft is a %s" .
            " which uses %s Kg of fuel on takeoff/landing, ". 
            " and %s Kg/Km while cruising; ".
            " %s%% of %s seats occupied," .
            " radiative forcing index is %s." .
            " Calculation: CO2 per passenger is %2d Kg," . 
            " all gasses (expressed as CO2 equivalent) is %2d Kg.",  
  $model["aircraft"], $model["fuel_for_takeoff"],  $model["fuel_cruising_per_km"],
  100 * $model["occupancy"],$model["capacity"], $model["radiative_forcing_index"],
  $co2, $co2_and_other_gasses);

  $emission->units = $node->reading ;   // this was previously set !
  $emission->co2 = $co2_and_other_gasses * _evaluate($node->adjustment) * $adjust ; // Kg
}


/**
 * Create a printable carbon footprint for a given period (usually a year)
 * based on on all the available carbon_stamps.
 *
 * @param $results a complicated structure containing all results of calculations
 * @param $json set to true if want table as raw data
 * @return a string containing a table that shows the carbon footprint.
 */


function carbon_calculate_present_results(&$results)
{
  $rows = array();
  $grand_total = 0 ;
  $numupdates = 0 ;
  $notes = array();
  $show_model = false ;


  // now go through the results and create a table
  // containing the results and then add the notes
  // to the end of the table.

  foreach ($results->emissions as $emission)
  {
    $result_as_text = null ;
    $column_notes = "" ;
    $showRow = false ;
    /*
     if (!isset($emission->select))
     {
      
     //      echo ("got row");
     $row = array( _format_date($emission->valuedate, "custom", "Y-m-d"),
     $emission->title, "benefic","","",$emission->toco2,"" );
     $rows[] = $row ;
     $numupdates++ ;
     continue ;
     }
     */

    if (!empty($emission->notes))
    {
      $notes[] = $emission->notes;
      $column_notes =  " [".count($notes)."]" ;
    }

    if (isset($emission->sourceError) || isset($emission->co2))
    {
      $showRow = true ;
    }
    if ($showRow)
    {
      $select = $emission->select ;
      $daterange = "" ;
      if ($select[0]->scope == 1)
      {
        if ($emission->start->enddate != $emission->end->enddate)
          $daterange = _format_date($emission->start->enddate, "custom", "Y-m-d")." and " ;
        $daterange .= _format_date($emission->end->enddate, "custom", "Y-m-d");
      }
      else
      {
        if (!empty($select[0]->startdate) && ($select[0]->startdate != $select[0]->enddate))
          $daterange = _format_date($select[0]->startdate, "custom", "Y-m-d")." to ";
        $daterange .= _format_date($select[0]->enddate, "custom", "Y-m-d");
        $emission->end = $select[0] ;
      }
      if (isset($emission->sourceError))
      {
        $units_col = $emission->sourceError ;
      }
      else   // ja 2-1-2008 replace select[0] with emission->end
      {
        if (isset($emission->units))
        $units = is_numeric($emission->units) ? sprintf("%.0f",$emission->units) : $emission->units ;
        else
        $units = $emission->end->amount ; // is this line used?

        $model = $show_model ? $emission->end->sourcemodel."/" : "" ;
        $units_col =  isset($units) ? $units." ".$emission->end->units." of " : "" ;

        $units_col .= l($model.$emission->end->sourcetitle, 'node/'.$emission->end->sourceid);
      }
      $row = array(
        $daterange,
        carbon_link_from_title($emission->end,$footprint)." ".$column_notes,
        $units_col);
      if ($results->enablekwh0)
        $row[] = sprintf("%.0f",$emission->kwh0);
      if ($results->enablekwh1)
        $row[] = sprintf("%.0f",$emission->kwh1);
      if ($results->enableco2)
        $row[] = empty($emission->error)  ? sprintf("%.0f",$emission->co2) : $emission->error;

      $row[] = carbon_format_edit_ops($emission->end, "destination=node/".$footprint->nid);
      $rows[] = $row ;
      $numupdates++ ;
    }
  }
  if (!empty($numupdates))
  {
    $out = "<h1 class='title'>" ;
    $out .= t("Emissions from ");
    $out .= substr(_format_date($results->start, "small"),0,10)." to " ;
    $out .= substr(_format_date($results->end, "small"),0,10) ;
    $out .= "</h1>" ;
    $row = array("total", null, null);
     
    if ($results->enablekwh0)
    $row[] = sprintf("%.0f",$results->grand_kwh0);

    if ($results->enablekwh1)
    $row[] = sprintf("%.0f",$results->grand_kwh1);

    if ($results->enableco2)
    $row[] = sprintf("%.2f",$results->grand_co2);

    // pass the footprint and see whether we are allowed to update
    // it rather than asking if we can crate a new stamp as the former
    // check makes sure we own the existing footprint.
    $row[] = carbon_format_add_new_op($result->account, "destination=node/".$results->account->nid);
    $rows[] = $row ;

    $header = array("date","title","source");
    if ($results->enablekwh0)
    $header[] = "renewable KwH" ;
    if ($results->enablekwh1)
    $header[] = "non renewable KwH" ;
    if ($results->enableco2)
    $header[] = "Kg of CO2" ;
    $header[] = "Operations";

    $out .= theme('table', $header, $rows);

    $i = 1 ; foreach ($notes as $note)
    {
      $out .= "<p class='link'>[".$i++."] ".$note."</p>" ;
    }
  }
  return $out ;
}


function carbon_calculate_empty_table($footprint)
{
  $row = array(t("There are no carbon stamps attached to this carbon footprint. <br/>" .
               "Click on add new to enter the first carbon stamp"), null);
  $row[] = carbon_format_add_new_op($footprint, "destination=node/".$footprint->nid);
  $out .= "<h1></h1>".theme('table', array("","",""), array($row));
  return $out ;
}

/*
 * PHP clone, backwards compatible
 */
function php4_clone($object)
{
  if (version_compare(phpversion(), '5.0') < 0)
    return $object;
  else
    return @clone($object);
}



?>
